from typing import Any, List

import pandas as pd
import structlog
from sqlalchemy.exc import NoResultFound
from sqlalchemy.pool import QueuePool
from sqlmodel import (
    Session,
    SQLModel,
    String,
    and_,
    create_engine,
    delete,
    func,
    select,
)
from sqlmodel.sql.expression import ColumnElement, SelectOfScalar

from no996.base.config import settings
from no996.base.date import Date

logger = structlog.get_logger(__name__)

engine = create_engine(
    str(settings.SQLALCHEMY_DATABASE_URI),
    pool_size=5,
    max_overflow=10,
    pool_pre_ping=True,
    pool_timeout=30,
    pool_recycle=3600,
    poolclass=QueuePool,
)


# https://github.com/fastapi/sqlmodel/issues/254
class DBMixin:
    def __init__(self, model: type[SQLModel], session: Session, show_log=True):
        self.model = model
        self.session = session
        self.show_log = show_log

    def get_model(self) -> type[SQLModel]:
        return self.model

    def get_pk_column(self) -> List[str]:
        primary_key_list = []
        for name, column in self.model.model_fields.items():
            if not hasattr(column, "primary_key"):
                return
            primary_key_status = column.primary_key
            if isinstance(primary_key_status, bool) and primary_key_status:
                primary_key_list.append(name)
        return primary_key_list 

    def compile(self, statement: ColumnElement | SelectOfScalar):
        if self.show_log:
            # 编译SQL查询语句，获取编译结果
            compiled = statement.compile()
            #
            # 获取参数化的值
            params = compiled.params

            logger.info(f"SQL: {compiled}")
            logger.info(f"Params: {params}")

    def first(self):
        statement = select(self.model)
        return self.session.exec(statement).first()

    def max_field(self, field: str):
        statement = select(func.max(getattr(self.model, field)))
        self.compile(statement)
        return self.session.exec(statement).first()

    def one_by_id(self, _id: int):
        obj = self.session.get(self.model, _id)
        return obj

    def first_by_field(self, field: str, value: Any):
        return self.first_by_fields({field: value})

    def one_by_field(self, field: str, value: Any):
        return self.one_by_fields({field: value})

    def first_by_fields(
        self,
        where_statement: dict | ColumnElement,
        order_by: ColumnElement | None = None,
    ):
        statement = select(self.model)

        if isinstance(where_statement, ColumnElement):
            statement = statement.where(where_statement)
        else:
            for key, value in where_statement.items():
                statement = statement.where(getattr(self.model, key) == value)

        if order_by is not None:
            statement = statement.order_by(order_by)
        self.compile(statement)
        return self.session.exec(statement).first()

    def one_by_fields(self, fields: dict):
        statement = select(self.model)
        for key, value in fields.items():
            statement = statement.where(getattr(self.model, key) == value)
        try:
            return self.session.exec(statement).one()
        except NoResultFound:
            logger.error(f"{self.model}: one_by_fields failed, NoResultFound")
            return None

    def all_by_field(self, field: str, value: Any):
        statement = select(self.model).where(getattr(self.model, field) == value)
        return self.session.exec(statement).all()

    def all_by_fields(
        self,
        where_statement: dict | ColumnElement | None = None,
        order_by: ColumnElement | None = None,
    ):
        statement = select(self.model)

        if where_statement is not None:
            if isinstance(where_statement, ColumnElement):
                statement = statement.where(where_statement)
            else:
                for key, value in where_statement.items():
                    statement = statement.where(getattr(self.model, key) == value)

        if order_by is not None:
            statement = statement.order_by(order_by)

        self.compile(statement)
        return self.session.exec(statement).all()

    def convert_without_saving(
        self, source: dict | SQLModel, update: dict | None = None
    ) -> SQLModel:
        return self.model.model_validate(source, update=update)

    def create(self, source: dict | SQLModel, update: dict | None = None):
        obj = self.convert_without_saving(source, update)
        self.save(obj)
        return obj

    def create_one_a_day(
        self,
        source: dict | SQLModel,
        update: dict | None = None,
        where_dict: dict = None,
    ) -> SQLModel | None:
        """
        创建一条记录，一天仅仅就一条，会将 created_at 作为like条件，如果已经存在则不创建
        """
        if where_dict is None:
            where_dict = {}

        created_at = Date.date_now(settings.DB_FMT)
        if "created_at" in where_dict:
            created_at = where_dict["created_at"]
        else:
            where_dict.update({"created_at": created_at})

        where_list = [func.cast(self.model.created_at, String).like(f"%{created_at}%")]

        for key, value in where_dict.items():
            if key != "created_at":
                where_list.append(getattr(self.model, key) == value)

        count = self.get_count(and_(*where_list))
        if count > 0:
            logger.error("当天仅能记录一条记录. 不得再次创建.")
            return None
        else:
            return self.create(source, update)  # Create

    def create_or_update(
        self, source: dict | SQLModel, update: dict | None = None
    ) -> SQLModel | None:
        obj = self.convert_without_saving(source, update)
        pk = self.model.__mapper__.primary_key_from_instance(obj)

        if pk[0] is not None:
            self.update(obj, where_dict=source)
            return self.one_by_fields(source)
        else:
            return self.create(obj)

    # https://github.com/tiangolo/sqlmodel/issues/494
    def get_count(self, where_statement: dict | ColumnElement) -> int:
        statement = select(self.model)

        if isinstance(where_statement, dict):
            for key, value in where_statement.items():
                statement = statement.where(getattr(self.model, key) == value)
        else:
            statement = statement.where(where_statement)

        q = statement

        count_q = (
            q.with_only_columns(func.count())
            .order_by(None)
            .select_from(q.get_final_froms()[0])
        )

        self.compile(statement)

        result = self.session.exec(count_q)
        return result.one()

    def save(self, data: SQLModel | list[SQLModel], commit=True) -> bool:
        self.session.add(data)
        if commit:
            self.session.commit()

    def update(self, update_dict: dict | SQLModel, where_dict: dict | SQLModel):
        statement = select(self.model)
        for key, value in where_dict.items():
            statement = statement.where(getattr(self.model, key) == value)
        self.compile(statement)
        fetched_model = self.session.exec(statement).all()

        if not fetched_model:
            return 0

        for model in fetched_model:
            if isinstance(update_dict, SQLModel):
                update_dict = update_dict.model_dump(exclude_unset=True)
            for key, value in update_dict.items():
                setattr(model, key, value)  # 更新获取到的模型，而不是传入的模型
            self.save(model, commit=False)

        self.session.commit()

        return len(fetched_model)

    def delete(self, where_statement: dict | ColumnElement = None):
        if where_statement is None:
            where_statement = {}
        statement = delete(self.model)

        if isinstance(where_statement, dict):
            for key, value in where_statement.items():
                statement = statement.where(and_(getattr(self.model, key) == value))
        else:
            statement = statement.where(where_statement)

        self.compile(statement)
        result = self.session.exec(statement)
        self.session.commit()
        return result.rowcount

    def all(self):
        statement = select(self.model)
        self.compile(statement)
        return self.session.exec(statement).all()

    def to_pandas(self) -> pd.DataFrame:
        records = self.all()
        return pd.json_normalize([r.model_dump() for r in records], sep="_")

    # https://github.com/tiangolo/sqlmodel/issues/215
    def model2df(self, objects: List[SQLModel], set_index: bool = True) -> pd.DataFrame:
        if len(objects) > 0:
            records = [obj.model_dump() for obj in objects]
            columns = list(objects[0].model_json_schema()["properties"].keys())
            df = pd.DataFrame.from_records(records, columns=columns)
            return df.set_index(columns[0]) if set_index else df
        return pd.DataFrame()

    def df_to_model(self, df: pd.DataFrame):
        # self.session.bulk_insert_mappings(self.model, df.to_dict("records"))
        # self.session.commit()

        batch_size = 1000  # 根据硬件调整（500-5000）
        records = df.to_dict("records")
        for i in range(0, len(records), batch_size):
            self.session.bulk_insert_mappings(
                self.model, 
                records[i:i+batch_size]
            )
            self.session.commit()  # 每次提交一个批次
            self.session.expire_all()  # 清除会话缓存防内存膨胀



from pydantic import AmqpDsn, PostgresDsn, computed_field
from pydantic_core import MultiHostUrl
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_file=".env", env_ignore_empty=True, extra="ignore"
    )

    TIME_ZONE: str

    DB_FMT: str
    DB_TIME_FMT: str

    POSTGRES_SERVER: str
    POSTGRES_DB: str
    POSTGRES_PORT: int
    POSTGRES_USER: str
    POSTGRES_PASSWORD: str

    @computed_field  # type: ignore[misc]
    @property
    def SQLALCHEMY_DATABASE_URI(self) -> PostgresDsn:
        return MultiHostUrl.build(
            scheme="postgresql+psycopg",
            username=self.POSTGRES_USER,
            password=self.POSTGRES_PASSWORD,
            host=self.POSTGRES_SERVER,
            port=self.POSTGRES_PORT,
            path=self.POSTGRES_DB,
        )

    AMQP_SERVER: str
    AMQP_PORT: int = 5432
    AMQP_USER: str
    AMQP_PASSWORD: str
    AMQP_VHOST: str = "/"

    @computed_field
    @property
    def AMQP_URI(self) -> AmqpDsn:
        return MultiHostUrl.build(
            scheme="amqp",
            username=self.AMQP_USER,
            password=self.AMQP_PASSWORD,
            host=self.AMQP_SERVER,
            port=self.AMQP_PORT,
            path=self.AMQP_VHOST,
        )

    # REDIS_SERVER: str
    # REDIS_PORT: int
    # REDIS_USER: str
    # REDIS_PASSWORD: str

    # @computed_field
    # @property
    # def REDIS_URI(self) -> RedisDsn:
    #     return MultiHostUrl.build(
    #         scheme="redis",
    #         host=self.REDIS_SERVER,
    #         port=self.REDIS_PORT,
    #         password=self.REDIS_PASSWORD,
    #     )

    # @computed_field
    # @property
    # def TASKIQ_BACKEND_URI(self) -> RedisDsn:
    #     return MultiHostUrl.build(
    #         scheme="redis",
    #         host=self.REDIS_SERVER,
    #         username="default",
    #         port=self.REDIS_PORT,
    #         password=self.REDIS_PASSWORD,
    #         path="0",
    #     )


settings = Settings()



import json
from typing import Any, Dict, List, Union

import aio_pika
import structlog
from aio_pika.abc import AbstractConnection
from aio_pika.pool import Pool

from no996.base.config import settings

logger = structlog.get_logger(__name__)


class MQ(object):
    _instance = None
    _connection_pool = None
    _channel_pools = {}
    _logger = logger

    def __new__(cls) -> "MQ":
        if cls._instance is None:
            cls._instance = super(MQ, cls).__new__(cls)
            cls._settings = settings
            cls._logger.info("MQ instance created")
        return cls._instance

    @classmethod
    async def get_connection_pool(cls) -> Pool:
        if cls._connection_pool is None:
            cls._connection_pool = Pool(
                cls._get_connection,
                max_size=5,  # Adjust pool size as needed
            )
            cls._logger.info("MQ connection pool established")
        return cls._connection_pool

    @classmethod
    async def _get_connection(cls) -> AbstractConnection:
        return await aio_pika.connect_robust(str(settings.AMQP_URI))

    @classmethod
    async def get_channel_pool(cls, connection: AbstractConnection) -> Pool:
        connection_id = id(connection)
        if connection_id not in cls._channel_pools:
            cls._channel_pools[connection_id] = Pool(
                connection.channel,
                max_size=10,  # Adjust channel pool size as needed
            )
        return cls._channel_pools[connection_id]

    @classmethod
    async def close(cls):
        if cls._channel_pools:
            for pool in cls._channel_pools.values():
                await pool.close()
            cls._channel_pools.clear()
        if cls._connection_pool is not None:
            await cls._connection_pool.close()
            cls._connection_pool = None
            cls._logger.info("MQ pools closed")

    async def publish_message(
        self, exchange_name: str, routing_key: str, message: Dict[str, Any]
    ):
        """Publish message to specified exchange and routing"""
        connection_pool = await self.get_connection_pool()
        async with connection_pool.acquire() as connection:
            channel_pool = await self.get_channel_pool(connection)
            async with channel_pool.acquire() as channel:
                # Declare exchange
                exchange = await channel.declare_exchange(
                    exchange_name, aio_pika.ExchangeType.TOPIC, durable=True
                )

                # Convert message to JSON and send
                message_body = json.dumps(message).encode()
                await exchange.publish(
                    aio_pika.Message(
                        body=message_body,
                        delivery_mode=aio_pika.DeliveryMode.PERSISTENT,
                    ),
                    routing_key=routing_key,
                )

                self._logger.info(
                    f"Published message to {exchange_name} with routing key {routing_key}"
                )

    async def consume_messages(
        self,
        queue_name: str,
        callback,
        routing_keys: Union[List[str], str] = [],
        exchange_name: str = None,
    ):
        """Consume messages from specified queue"""
        connection_pool = await self.get_connection_pool()
        async with connection_pool.acquire() as connection:
            channel_pool = await self.get_channel_pool(connection)
            async with channel_pool.acquire() as channel:
                await channel.set_qos(prefetch_count=1)

                # Handle exchange and routing key binding if specified
                if exchange_name and routing_keys:
                    exchange = await channel.declare_exchange(
                        exchange_name, aio_pika.ExchangeType.TOPIC, durable=True
                    )

                    queue = await channel.declare_queue(queue_name, durable=True)

                    if isinstance(routing_keys, str):
                        routing_keys = [routing_keys]

                    for routing_key in routing_keys:
                        await queue.bind(exchange, routing_key)
                else:
                    queue = await channel.declare_queue(queue_name, durable=True)

                async with queue.iterator() as queue_iter:
                    async for message in queue_iter:
                        # 当代码块执行完成后，会自动确认(ack)消息
                        # 如果代码块执行过程中抛出异常，消息会被自动拒绝(nack)
                        async with message.process():
                            logger.info(f"Received message: {message.body.decode()}")
                            data = json.loads(message.body.decode())
                            await callback(data)



import os


class File(object):
    @staticmethod
    # 递归搜索指定文件夹下所有文件
    def walk(path, callback=None):
        file_list = []
        for root, dirs, files in os.walk(path):
            for file in files:
                abs_path = os.path.join(root, file)
                if callback:
                    abs_path = callback(abs_path)
                if abs_path:
                    file_list.append(abs_path)
        return file_list



import arrow
import pytz
import structlog

from no996.base.config import settings

logger = structlog.get_logger(__name__)


class Date:
    @staticmethod
    def date_fmt(date: str | arrow.Arrow, fmt: str) -> str:
        if isinstance(date, str):
            date = arrow.get(date)
        return date.format(fmt=fmt)

    @staticmethod
    def str2date(
        date: str, fmt: str, district: str | None = settings.TIME_ZONE
    ) -> arrow.Arrow:
        """字符串转日期

        Args:
            date (str): 日期字符串
            fmt (str): 日期格式
            district (Optional[str], optional): 时区. Defaults to 'Asia/Shanghai'.

        Returns:
            arrow.Arrow: 日期对象
        """
        if district.lower() != "utc":
            return arrow.get(date, fmt).replace(tzinfo=district)

        return arrow.get(date, fmt)

    @staticmethod
    def date_now(
        fmt: str | None = None, district: str | None = settings.TIME_ZONE
    ) -> arrow.Arrow | str:
        """获取当前区域的时间

        Args:
            fmt (Optional[str], optional): _description_. Defaults to None.
            district (Optional[str], optional): _description_. Defaults to 'Asia/Shanghai'.

        Returns:
            Union[arrow.Arrow, str]: _description_
        """

        current_time = arrow.utcnow()
        if district.lower() != "utc":
            timezone = pytz.timezone(district)
            current_time = current_time.to(timezone)

        if not fmt:
            return current_time
        return current_time.format(fmt)

    @staticmethod
    def date_range(
        start_date: str | arrow.Arrow,
        end_date: str | arrow.Arrow,
        fmt: str | None = None,
    ) -> list[arrow.Arrow] | list[str]:
        """获取日期区间内所有的日期


        Args:
            start_date (Union[str, arrow.Arrow]): 起始日期
            end_date (Union[str, arrow.Arrow]): 结束日期
            fmt (Optional[str], optional): 日期格式. Defaults to None.

        Returns:
            Union[List[arrow.Arrow], List[str]]: 返回类型
        """
        if isinstance(start_date, str):
            start_date = arrow.get(start_date)
        if isinstance(end_date, str):
            end_date = arrow.get(end_date)

        if not fmt:
            return [
                arrow.get(i) for i in arrow.Arrow.range("day", start_date, end_date)
            ]
        return [
            arrow.get(i).format(fmt)
            for i in arrow.Arrow.range("day", start_date, end_date)
        ]

    @staticmethod
    def get_dates_before_or_after(
        target_date: str | arrow.Arrow,
        date_list: list[str],
        fmt: str | None,
        before: bool = False,
        sort_dates: bool = False,
    ) -> list[str]:
        target_date_obj = (
            arrow.get(target_date) if isinstance(target_date, str) else target_date
        )
        date_obj_list = [arrow.get(date) for date in date_list]
        filtered_dates = filter(
            lambda date: (date <= target_date_obj and before)
            or (date > target_date_obj and not before),
            date_obj_list,
        )
        if fmt:
            filtered_dates = [
                Date.date_fmt(date_obj, fmt=fmt) for date_obj in filtered_dates
            ]

        sorted_dates = sorted(filtered_dates) if sort_dates else list(filtered_dates)

        return sorted_dates

    @staticmethod
    def date_cal(
        date: str | arrow.Arrow, days: int, fmt: str = settings.DB_FMT
    ) -> str | arrow.Arrow:
        """日期计算器

        Args:
            date (Union[str, arrow.Arrow]): 日期
            days (int): 天数
            fmt (str): 日期格式. Defaults to DB_FMT.

        Returns:
            Union[str, arrow.Arrow]: 计算结果
        """
        current_date = arrow.get(date)
        new_date = current_date.shift(days=days)
        if isinstance(date, str):
            return new_date.format(fmt)
        return new_date

    @staticmethod
    def date_sep(
        start_date: str | arrow.Arrow,
        end_date: str | arrow.Arrow,
        fmt: str = settings.DB_FMT,
        max_value: int = 300,
    ) -> list[str | arrow.Arrow]:
        """日期分割器

        Args:
            start_date (Union[str, arrow.Arrow]): 起始日期
            end_date (Union[str, arrow.Arrow]): 当前日期
            fmt (str): 返回日期格式
            max_value (int): 分割间隔

        Yields:
            list: 日期列表
        """
        __start_date = start_date
        if isinstance(__start_date, str):
            start_date = arrow.get(start_date)
            end_date = arrow.get(end_date)

        while start_date <= end_date:
            current_date = start_date.shift(days=max_value)

            if current_date > end_date:
                current_date = end_date
                current_date_value = current_date.format(fmt)

            if isinstance(__start_date, str):
                start_date_value = start_date.format(fmt)
                current_date_value = current_date.format(fmt)
                yield [start_date_value, current_date_value]
            else:
                yield [start_date, current_date]

            start_date = current_date.shift(days=1)

    @staticmethod
    def date_compare(
        start_date: str | arrow.Arrow, end_date: str | arrow.Arrow
    ) -> bool:
        """日期比较器

        Args:
            start_date (Union[str, arrow.Arrow]): 起始日期
            end_date (Union[str, arrow.Arrow]): 结束日期

        Returns:
            bool: 布尔值
        """
        if isinstance(start_date, str):
            start_date = arrow.get(start_date)
        if isinstance(end_date, str):
            end_date = arrow.get(end_date)

        return end_date > start_date

    @staticmethod
    def date_quarter_list(
        start_date: str | arrow.Arrow, end_date: str | arrow.Arrow, fmt: str
    ) -> list[str]:
        """获取起始和结束之前的比当前日期小的季度日期

        Returns:
            _type_: _description_
        """
        if isinstance(start_date, str):
            start_date = arrow.get(start_date)
        if isinstance(end_date, str):
            end_date = arrow.get(end_date)

        date_list = []

        start_year = start_date.format("YYYY")
        end_year = end_date.format("YYYY")
        quarter_end_list = ["0331", "0630", "0930", "1231"]
        for year in range(int(start_year), int(end_year) + 1):
            for quarter_date in quarter_end_list:
                quarter_date = f"{year}{quarter_date}"
                if not Date.date_compare(quarter_date, Date.date_now()):
                    break
                quarter_date = arrow.get(quarter_date)
                if fmt:
                    quarter_date = quarter_date.format(fmt)
                date_list.append(quarter_date)
        return date_list

    @staticmethod
    def date_month_list(
        start_date: str | arrow.Arrow,
        end_date: str | arrow.Arrow,
        fmt: str | None = None,
    ) -> list[arrow.Arrow] | list[str]:
        """获取月份列表

        Args:
            start_date (Union[str, arrow.Arrow]): 起始日期
            end_date (Union[str, arrow.Arrow]): 结束日期
            fmt (Optional[str], optional): 日期格式. Defaults to None.

        Returns:
            Union[List[arrow.Arrow], List[str]]: 日期列表
        """
        if isinstance(start_date, str):
            start_date = arrow.get(start_date)
        if isinstance(end_date, str):
            end_date = arrow.get(end_date)

        month_list = []
        if fmt:
            month_list = [
                date.format(fmt)
                for date in arrow.Arrow.range("month", start_date, end_date)
            ]

        else:
            month_list = list(arrow.Arrow.range("month", start_date, end_date))

        return month_list

    @staticmethod
    # 判断日期是否是周末
    def is_weekend(date: str | arrow.Arrow) -> bool:
        """判断日期是否是周末

        Args:
            date (Union[str, arrow.Arrow]): 日期

        Returns:
            bool: 布尔值
        """
        if isinstance(date, str):
            date = arrow.get(date)
        return date.weekday() in [5, 6]

    @staticmethod
    def get_last_weekend(
        date: str | arrow.Arrow, weeks=-1, weekday=4, fmt: str | None = None
    ) -> str | arrow.Arrow:
        """
        获取当前周数前后的指定星期几

        Args:
            date (Union[str, arrow.Arrow]): The date to calculate the last weekend from.
            weeks (int, optional): The number of weeks to shift the date. Defaults to -1.
            weekday (int, optional): The weekday to consider as the weekend. Defaults to 4 (Friday).
            fmt (str, optional): The format to return the date in. Defaults to None.

        Returns:
            str: The date of the last weekend.

        """
        if isinstance(date, str):
            date = arrow.get(date)
        if date.weekday() in [5, 6]:
            date = date.shift(weeks=weeks, weekday=weekday)

        if fmt:
            return date.format(fmt)
        return date

    @staticmethod
    def get_last_date_or_today(
        provided_date: str | arrow.Arrow, fmt: str | None = None
    ) -> str | arrow.Arrow:
        """
        提供一个日期 比如 19910101 如果该日期小于今年，则要返回结束日期为 当年年底的日期 19911231 否则，返回今天日期
        """
        # 获取今天的日期
        today = arrow.now()

        if isinstance(provided_date, str):
            # 转换为 Arrow 对象
            provided_arrow = arrow.get(provided_date)
        else:
            provided_arrow = provided_date

        # 判断提供的日期是否小于今年
        if provided_arrow.year < today.year:
            start_date = arrow.get(provided_arrow.year, 1, 1)  # 获取年初日期
            end_date = start_date.shift(years=1, days=-1)  # 加1年再减一天
            result_date = end_date.format(fmt)
        else:
            result_date = today.format(fmt)
        return result_date



import time
from functools import wraps

import structlog

logger = structlog.get_logger(__name__)


def timer_decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()  # 获取函数开始运行时间
        result = func(*args, **kwargs)  # 执行函数
        end_time = time.time()  # 获取函数结束运行时间
        elapsed_time = (end_time - start_time) * 1000  # 计算函数运行时间（单位：毫秒）

        cost = f"{elapsed_time:.3f}"

        msg = f"[{func.__name__}] 函数运行时间为 {cost} 毫秒"
        if elapsed_time < 3000:
            logger.info(msg)
        else:
            logger.warning(msg)
        return result

    return wrapper
